/*
 * stack.h
 *
 *  Created on: 18 Oct 2011
 *  Copyright 2011 Nicolas Melot
 *
 * This file is part of TDDD56.
 *
 *     TDDD56 is free software: you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation, either version 3 of the License, or
 *     (at your option) any later version.
 *
 *     TDDD56 is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with TDDD56. If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include <stdlib.h>
#include <pthread.h>

#ifndef STACK_H
#define STACK_H

struct node {
  struct node* next;
  int value;
};
typedef struct node node_t;

struct stack
{
  node_t* head;
  int size;
  int change_this_member;
};
typedef struct stack stack_t;

void stack_init(stack_t* stack, int no_of_nodes) ;
void stack_push(stack_t *stack, stack_t *free_stack, int data);
int stack_pop(stack_t *stack, stack_t *free_stack);

#if NON_BLOCKING == 1
void stack_push_ABA(stack_t *stack, stack_t *free_stack, int id, int* flag);
int stack_pop_ABA(stack_t *stack, stack_t *free_stack, int id, int* flag);
#endif
void stack_push_free(stack_t *stack, node_t* node);
node_t* stack_pop_free(stack_t *stack);

void free_nodes(stack_t* stack);
void print_sizes(stack_t* stack);
node_t* allocate_node(stack_t* stack);
/* Use this to check if your stack is in a consistent state from time to time */
int stack_check(stack_t *stack, int gt);
int stack_size(stack_t* stack);
#endif /* STACK_H */
