// Simple CUDA example by Ingemar Ragnemalm 2009. Simplest possible?
// Assigns every element in an array with its index.

// nvcc simple.cu -L /usr/local/cuda/lib -lcudart -o simple

#include <stdio.h>
#include <cmath>

__global__
void add(float* result, float *c, float *d, int N)
{
	int idx = blockIdx.x * blockDim.x + threadIdx.x;
	int idy = blockIdx.y * blockDim.y + threadIdx.y;
	int index = idx + idy * N;

	if (idx < N && idy < N) {
			result[index] = c[index] + d[index];
	}

}

int main(int argc, char *argv[])
{
	char *argument = argv[argc-1];
  int N = atoi(argument);
	int blocksize = 32; //threads
	int gridsize = 32; //blocks
  printf("Running on GPU with no of elements: %i \n",N);
	float *a = new float[N*N];
	float *b = new float[N*N];
	float *result = new float[N*N];
	float theTime = 0.0;
	cudaEvent_t startEvent, finEvent;


	for (int i = 0; i < N; i++)
		for (int j = 0; j < N; j++)
		{
			a[i+j*N] = 10 + i;
			b[i+j*N] = (float)j / N;
			result[i+j*N] = 0.0;
		}
	float *c;
	float *d;
	float *result_cuda;
	const int size = N*N*sizeof(float);

	cudaMalloc( (void**)&c, size );
	cudaMalloc( (void**)&d, size );
	cudaMalloc( (void**)&result_cuda, size );
	dim3 dimBlock( blocksize, blocksize );
	dim3 dimGrid( gridsize, gridsize );
	cudaMemcpy( c, a, size, cudaMemcpyHostToDevice );
	cudaMemcpy( d, b, size, cudaMemcpyHostToDevice );

  cudaEventCreate(&startEvent);
	cudaEventCreate(&finEvent);
	cudaEventRecord(startEvent, 0);
	add<<<dimGrid, dimBlock>>>(result_cuda, c, d, N);
	cudaThreadSynchronize();
	cudaEventRecord(finEvent, 0);
	cudaEventSynchronize(finEvent);

	cudaEventElapsedTime(&theTime, startEvent, finEvent);

	cudaMemcpy( result, result_cuda, size, cudaMemcpyDeviceToHost );
	cudaFree( c );
	cudaFree( d );
	cudaFree( result_cuda );

	for (int i = 0; i < N; i++)
	{
		for (int j = 0; j < N; j++)
		{
			printf("%0.2f ", result[i+j*N]);
		}
		printf("\n");
	}
	cudaError_t err = cudaGetLastError();

	 if (err != cudaSuccess)
			 printf("Error: %s\n", cudaGetErrorString(err));

	printf("%s%f\n", "Time elapsed (ms): ", theTime);
	cudaDeviceProp prop;
  cudaGetDeviceProperties(&prop, 0);
  printf("Device Number: %d\n", 0);
  printf("  Device name: %s\n", prop.name);
  printf("  Memory Clock Rate (KHz): %d\n",
         prop.memoryClockRate);
  printf("  Memory Bus Width (bits): %d\n",
         prop.memoryBusWidth);
  printf("  Peak Memory Bandwidth (GB/s): %f\n",
         2.0*prop.memoryClockRate*(prop.memoryBusWidth/8)/1.0e6);
  printf("  Max threads per block: %i\n",
         prop.maxThreadsPerBlock);
	printf("  Max threads dim: %i x %i x %i\n",
         prop.maxThreadsDim[0],prop.maxThreadsDim[1],prop.maxThreadsDim[2] );
  printf("  Max grid size: %i x %i x %i\n",
         prop.maxGridSize[0],prop.maxGridSize[1],prop.maxGridSize[2] );
	delete[] a;
	delete[] b;
	printf("done\n");
	return EXIT_SUCCESS;
}
