// Lab 5, image filters with CUDA.

// Compile with a command-line similar to Lab 4:
// nvcc filter.cu -c -arch=sm_30 -o filter.o
// g++ filter.o milli.c readppm.c -lGL -lm -lcuda -lcudart -L/usr/local/cuda/lib -lglut -o filter
// or (multicore lab)
// nvcc filter.cu -c -arch=sm_20 -o filter.o
// g++ filter.o milli.c readppm.c -lGL -lm -lcuda -L/usr/local/cuda/lib64 -lcudart -lglut -o filter

// 2017-11-27: Early pre-release, dubbed "beta".
// 2017-12-03: First official version! Brand new lab 5 based on the old lab 6.
// Better variable names, better prepared for some lab tasks. More changes may come
// but I call this version 1.0b2.
// 2017-12-04: Two fixes: Added command-lines (above), fixed a bug in computeImages
// that allocated too much memory. b3
// 2017-12-04: More fixes: Tightened up the kernel with edge clamping.
// Less code, nicer result (no borders). Cleaned up some messed up X and Y. b4

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#ifdef __APPLE__
  #include <GLUT/glut.h>
  #include <OpenGL/gl.h>
#else
  #include <GL/glut.h>
#endif
#include "readppm.h"
#include "milli.h"

// Use these for setting shared memory size.
#define maxKernelSizeX 10
#define maxKernelSizeY 10

// unsigned char median_kernel(skepu::Region2D<unsigned char> image, size_t elemPerPx)
/*__global__ void median_filter(unsigned char *image, unsigned char *out, const unsigned int imagesizex, const unsigned int imagesizey, const int kernelsizex, const int kernelsizey)
{
  __shared__ unsigned char patch[(tileSize*3+2*maxKernelSizeX*3)*(tileSize+2*maxKernelSizeY)];
  // map from blockIdx to pixel position
	int g_x = blockIdx.x * tileSize + threadIdx.x*3 - kernelsizex*3;
  int g_y = blockIdx.y * tileSize + threadIdx.y - kernelsizey;
  // Use max and min to avoid branching!
  g_x = min(max(g_x, 0), imagesizex*3-1);
  g_y = min(max(g_y, 0), imagesizey-1);

  int g_i = imagesizex*g_y*3 + g_x;
  int l_i = threadIdx.x*3 + blockDim.x*3*threadIdx.y;

  int l_x = threadIdx.x*3;
  int l_y = threadIdx.y;

  patch[l_i+0] = image[g_i+0];
  patch[l_i+1] = image[g_i+1];
  patch[l_i+2] = image[g_i+2];

  __syncthreads();
  // Make local copy for each thread
  const int size = (maxKernelSizeX*2+1)*(maxKernelSizeY*2+1);
  unsigned char r[size];
  unsigned char g[size];
  unsigned char b[size];
  int k = (2*kernelsizex+1)*(2*kernelsizey+1);
  unsigned char temp;

  int j = 0;
  int dx, dy;
  if((l_x >= (kernelsizex*3)) && (l_x < ((blockDim.x-(kernelsizex*3)))) &&
     (l_y >= kernelsizey) && (l_y < (blockDim.y-kernelsizey))) {
   	 for(dy=-kernelsizey;dy<=kernelsizey;dy++)
     {
     		for(dx=-kernelsizex;dx<=kernelsizex;dx++)
     		{
           r[j] = patch[(l_i+(dy*blockDim.x*3)+dx*3)+0];
           g[j] = patch[(l_i+(dy*blockDim.x*3)+dx*3)+1];
           b[j] = patch[(l_i+(dy*blockDim.x*3)+dx*3)+2];
           j++;
     		}
    }

    //Bubblesort
    for (int i = 0; i < k-1; i++){
      for (int j = 0; j < k-i-1; j++) {
        if (r[j] > r[j+1]) {
           temp = r[j];
           r[j] = r[j+1];
           r[j+1] = temp;
        }
        if (g[j] > g[j+1]) {
           temp = g[j];
           g[j] = g[j+1];
           g[j+1] = temp;
        }
        if (b[j] > b[j+1]) {
           temp = b[j];
           b[j] = b[j+1];
           b[j+1] = temp;
        }
      }
    }
    unsigned int median = (k-1)/2;


    out[g_i+0] = r[median];
    out[g_i+1] = g[median];
    out[g_i+2] = b[median];
  }
  /*


  */
  /*
    out[g_i+0] = patch[l_i];
    out[g_i+1] = patch[l_i+1];
    out[g_i+2] = patch[l_i+2];

}*/
__global__ void gaussian_filter(unsigned char *image, unsigned char *out, const unsigned int imagesizex, const unsigned int imagesizey, const int kernelsizex, const int kernelsizey, unsigned int *stencil)
{
  //__shared__ unsigned char patch[(32*3+(2*maxKernelSizeX*3))*(32+2*maxKernelSizeY)];
  const int tile = blockDim.x - 2*maxKernelSizeX;
  __shared__ unsigned char patch[32*3*32];
  // map from blockIdx to pixel position
	int g_x = blockIdx.x * tile + threadIdx.x-kernelsizex;
  int g_y = blockIdx.y * tile + threadIdx.y-kernelsizey;
  // Use max and min to avoid branching!
  g_x = min(max(g_x, 0), imagesizex-1);
  g_y = min(max(g_y, 0), imagesizey-1);

  int g_i = imagesizex*g_y + g_x;

  int l_i = threadIdx.x + blockDim.x*threadIdx.y;
  int l_x = threadIdx.x;
  int l_y = threadIdx.y;

  //(y*imagesizex+x)*3+0

  patch[l_i*3+0] = image[g_i*3+0];
  patch[l_i*3+1] = image[g_i*3+1];
  patch[l_i*3+2] = image[g_i*3+2];

  __syncthreads();

  if((l_x >= kernelsizex) && (l_x < (blockDim.x-kernelsizex)) &&
     (l_y >= kernelsizey) && (l_y < (blockDim.y-kernelsizey))) {

     unsigned int sumx, sumy, sumz;
     int dy, dx;
     sumx=0;sumy=0;sumz=0;
     int j = 0;
   	 for(dy=-kernelsizey;dy<=kernelsizey;dy++)
     {
     		for(dx=-kernelsizex;dx<=kernelsizex;dx++)
     		{

           sumx += patch[(l_i+(dy*blockDim.x)+dx)*3+0]*stencil[j];
           sumy += patch[(l_i+(dy*blockDim.x)+dx)*3+1]*stencil[j];
           sumz += patch[(l_i+(dy*blockDim.x)+dx)*3+2]*stencil[j];

           j++;
     		}
    }

    out[g_i*3+0] = sumx/16;
    out[g_i*3+1] = sumy/16;
    out[g_i*3+2] = sumz/16;

  }
}
  __global__ void median_filter(unsigned char *image, unsigned char *out, const unsigned int imagesizex, const unsigned int imagesizey, const int kernelsizex, const int kernelsizey)
  {
    const int tile = blockDim.x - 2*maxKernelSizeX;
    __shared__ unsigned char patch[32*3*32];
    // map from blockIdx to pixel position
    int g_x = blockIdx.x * tile + threadIdx.x-kernelsizex;
    int g_y = blockIdx.y * tile + threadIdx.y-kernelsizey;
    // Use max and min to avoid branching!
    g_x = min(max(g_x, 0), imagesizex-1);
    g_y = min(max(g_y, 0), imagesizey-1);

    int g_i = imagesizex*g_y + g_x;

    int l_i = threadIdx.x + blockDim.x*threadIdx.y;
    int l_x = threadIdx.x;
    int l_y = threadIdx.y;

    //(y*imagesizex+x)*3+0

    patch[l_i*3+0] = image[g_i*3+0];
    patch[l_i*3+1] = image[g_i*3+1];
    patch[l_i*3+2] = image[g_i*3+2];

    __syncthreads();
    // Make local copy for each thread
    const int size = (maxKernelSizeX*2+1)*(maxKernelSizeY*2+1);

    unsigned char r[size];
    unsigned char g[size];
    unsigned char b[size];
    int k = (2*kernelsizex+1)*(2*kernelsizey+1);
    unsigned char temp;

    int j = 0;
    int dx, dy;

    if((l_x >= (kernelsizex)) && (l_x < ((blockDim.x-(kernelsizex)))) &&
       (l_y >= kernelsizey) && (l_y < (blockDim.y-kernelsizey))) {
     	 for(dy=-kernelsizey;dy<=kernelsizey;dy++)
       {
       		for(dx=-kernelsizex;dx<=kernelsizex;dx++)
       		{
             r[j] = patch[(l_i+(dy*blockDim.x)+dx)*3+0];
             g[j] = patch[(l_i+(dy*blockDim.x)+dx)*3+1];
             b[j] = patch[(l_i+(dy*blockDim.x)+dx)*3+2];
             j++;
       		}
      }

      //Bubblesort
      for (int i = 0; i < k-1; i++){
        for (int j = 0; j < k-i-1; j++) {
          if (r[j] > r[j+1]) {
             temp = r[j];
             r[j] = r[j+1];
             r[j+1] = temp;
          }
          if (g[j] > g[j+1]) {
             temp = g[j];
             g[j] = g[j+1];
             g[j+1] = temp;
          }
          if (b[j] > b[j+1]) {
             temp = b[j];
             b[j] = b[j+1];
             b[j+1] = temp;
          }
        }
      }
      unsigned int median = (k-1)/2;


      out[g_i*3+0] = r[median];
      out[g_i*3+1] = g[median];
      out[g_i*3+2] = b[median];
    }
}
__global__ void filter(unsigned char *image, unsigned char *out, const unsigned int imagesizex, const unsigned int imagesizey, const int kernelsizex, const int kernelsizey)
{
  const int tile = blockDim.x - 2*maxKernelSizeX;
  __shared__ unsigned char patch[32*3*32];
  // map from blockIdx to pixel position
	int g_x = blockIdx.x * tile + threadIdx.x-kernelsizex;
  int g_y = blockIdx.y * tile + threadIdx.y-kernelsizey;
  // Use max and min to avoid branching!
  g_x = min(max(g_x, 0), imagesizex-1);
  g_y = min(max(g_y, 0), imagesizey-1);

  int g_i = imagesizex*g_y + g_x;

  int l_i = threadIdx.x + blockDim.x*threadIdx.y;
  int l_x = threadIdx.x;
  int l_y = threadIdx.y;

  //(y*imagesizex+x)*3+0

  patch[l_i*3+0] = image[g_i*3+0];
  patch[l_i*3+1] = image[g_i*3+1];
  patch[l_i*3+2] = image[g_i*3+2];

  __syncthreads();

  if((l_x >= kernelsizex) && (l_x < (blockDim.x-kernelsizex)) &&
     (l_y >= kernelsizey) && (l_y < (blockDim.y-kernelsizey))) {

     unsigned int sumx, sumy, sumz;
     int dy, dx;
     sumx=0;sumy=0;sumz=0;

   	 for(dy=-kernelsizey;dy<=kernelsizey;dy++)
     {
     		for(dx=-kernelsizex;dx<=kernelsizex;dx++)
     		{

           sumx += patch[(l_i+(dy*blockDim.x)+dx)*3+0];
           sumy += patch[(l_i+(dy*blockDim.x)+dx)*3+1];
           sumz += patch[(l_i+(dy*blockDim.x)+dx)*3+2];

           //patch[l_i+(dy*blockDim.x*3)+(dx*3)+0] = 255;
           //patch[l_i+(dy*blockDim.x*3)+(dx*3)+1] = 255;
           //patch[l_i+(dy*blockDim.x*3)+(dx*3)+2] = 255;
     		}
    }

    int divby = (2*kernelsizex+1)*(2*kernelsizey+1); // Works for box filters only!

    out[g_i*3+0] = sumx/divby;
    out[g_i*3+1] = sumy/divby;
    out[g_i*3+2] = sumz/divby;
    //out[g_i*3+0] = patch[l_i*3+0];
    //out[g_i*3+1] = patch[l_i*3+1];
    //out[g_i*3+2] = patch[l_i*3+2];
    //out[g_i+0] = image[g_i+0];
    //out[g_i+1] = image[g_i+1];
    //out[g_i+2] = image[g_i+2];

  }


}

// Global variables for image data

unsigned char *image, *pixels, *dev_bitmap, *dev_input, *dev_sep;
unsigned int imagesizey, imagesizex; // Image size
unsigned int stencil[5] = {1,4,6,4,1};
unsigned int *stencilcuda;

////////////////////////////////////////////////////////////////////////////////
// main computation function
////////////////////////////////////////////////////////////////////////////////
void computeImages(int kernelsizex, int kernelsizey)
{
	if (kernelsizex > maxKernelSizeX || kernelsizey > maxKernelSizeY)
	{
		printf("Kernel size out of bounds!\n");
		return;
	}

	pixels = (unsigned char *) malloc(imagesizex*imagesizey*3);
	cudaMalloc( (void**)&dev_input, imagesizex*imagesizey*3);
	cudaMemcpy( dev_input, image, imagesizey*imagesizex*3, cudaMemcpyHostToDevice );
	cudaMalloc( (void**)&dev_bitmap, imagesizex*imagesizey*3);
  cudaMalloc( (void**)&dev_sep, imagesizex*imagesizey*3);


  //dim3 blockSize(tileSize*3+(2*kernelsizex*3), tileSize+(2*kernelsizey));
  dim3 blockSize(32, 32);
  dim3 grid(imagesizex/maxKernelSizeX,imagesizey/maxKernelSizeY);
  //dim3 grid((imagesizex + blockSize.x - 1)/blockSize.x, (imagesizey + blockSize.y - 1)/blockSize.y);

  //Task 1
	filter<<<grid,blockSize>>>(dev_input, dev_bitmap, imagesizex, imagesizey, kernelsizex, kernelsizey);

  //Task 2
  //filter<<<grid,blockSize>>>(dev_input, dev_sep, imagesizex, imagesizey, kernelsizex, kernelsizey);
  //cudaDeviceSynchronize();
  //filter<<<grid,blockSize>>>(dev_sep, dev_bitmap, imagesizex, imagesizey, kernelsizey, kernelsizex);

  //Task 3 Gaussian
  /*cudaMalloc( (void**)&stencilcuda, 5*sizeof(unsigned int));
  cudaMemcpy( stencilcuda, stencil, 5*sizeof(unsigned int), cudaMemcpyHostToDevice);
  gaussian_filter<<<grid,blockSize>>>(dev_input, dev_bitmap, imagesizex, imagesizey, kernelsizex, kernelsizey, stencilcuda);
  cudaDeviceSynchronize();
  gaussian_filter<<<grid,blockSize>>>(dev_input, dev_bitmap, imagesizex, imagesizey, kernelsizey, kernelsizex, stencilcuda);*/
  //Task 4 Median
  //median_filter<<<grid,blockSize>>>(dev_input, dev_bitmap, imagesizex, imagesizey, kernelsizex, kernelsizey);
	cudaThreadSynchronize();
//	Check for errors!
    cudaError_t err = cudaGetLastError();
    if (err != cudaSuccess)
        printf("Error: %s\n", cudaGetErrorString(err));
	cudaMemcpy( pixels, dev_bitmap, imagesizey*imagesizex*3, cudaMemcpyDeviceToHost );
	cudaFree( dev_bitmap );
	cudaFree( dev_input );
  cudaFree(stencilcuda);
}

// Display images
void Draw()
{
// Dump the whole picture onto the screen.
	glClearColor( 0.0, 0.0, 0.0, 1.0 );
	glClear( GL_COLOR_BUFFER_BIT );

	if (imagesizey >= imagesizex)
	{ // Not wide - probably square. Original left, result right.
		glRasterPos2f(-1, -1);
		glDrawPixels( imagesizex, imagesizey, GL_RGB, GL_UNSIGNED_BYTE, image );
		glRasterPos2i(0, -1);
		glDrawPixels( imagesizex, imagesizey, GL_RGB, GL_UNSIGNED_BYTE,  pixels);
	}
	else
	{ // Wide image! Original on top, result below.
		glRasterPos2f(-1, -1);
		glDrawPixels( imagesizex, imagesizey, GL_RGB, GL_UNSIGNED_BYTE, pixels );
		glRasterPos2i(-1, 0);
		glDrawPixels( imagesizex, imagesizey, GL_RGB, GL_UNSIGNED_BYTE, image );
	}
	glFlush();
}

// Main program, inits
int main( int argc, char** argv)
{
	glutInit(&argc, argv);
	glutInitDisplayMode( GLUT_SINGLE | GLUT_RGBA );

	if (argc > 1)
		image = readppm(argv[1], (int *)&imagesizex, (int *)&imagesizey);
	else
		image = readppm((char *)"maskros512.ppm", (int *)&imagesizex, (int *)&imagesizey);

	if (imagesizey >= imagesizex)
		glutInitWindowSize( imagesizex*2, imagesizey );
	else
		glutInitWindowSize( imagesizex, imagesizey*2 );
	glutCreateWindow("Lab 5");
	glutDisplayFunc(Draw);

	ResetMilli();

  //Task 1,4
	//computeImages(10, 10);
  //Task 2
  //computeImages(10, 0);
  //computeImages(0, 10);
  //Task 3
  computeImages(3, 3);

  printf("%s%f\n","Measured time (ms): ", (float)GetMicroseconds()*0.001);
// You can save the result to a file like this:
	//writeppm("outgaussian.ppm", imagesizey, imagesizex, pixels);

	glutMainLoop();
	return 0;
}
